// effects.c
// LiVES (lives-exe)
// (c) G. Finch 2003
// Released under the GPL 2.0
// see file ../COPYING for licensing details

#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>

#include "main.h"
#include "effects.h"

//////////// Effects ////////////////


static gchar effectname[256];



//// some pixel utils

static inline guint 
calc_luma (guchar *pixel) {
  return (guint)(pixel[0]*30+pixel[1]*59+pixel[2]*11);
}

static inline void 
make_black (guchar *pixel) {
  pixel[0]=pixel[1]=pixel[2]=(guchar)0;
}

static inline void 
make_white (guchar *pixel) {
  pixel[0]=pixel[1]=pixel[2]=(guchar)255;
}


void 
nine_fill (guchar *new_data, gint i, gint j, guchar *old_data, gint x, gint y, gint rowstride) {
  // fill nine pixels with the centre colour
  new_data[i-rowstride+j-3]=new_data[i-rowstride]=new_data[i-rowstride+3]=new_data[i+j-3]=new_data[i+j]=new_data[i+j+3]=new_data[i+rowstride+j-3]=new_data[i+rowstride+j]=new_data[i+rowstride+j+3]=old_data[x+y];
  new_data[i-rowstride+j-2]=new_data[i-rowstride+1]=new_data[i-rowstride+4]=new_data[i+j-2]=new_data[i+j+1]=new_data[i+j+4]=new_data[i+rowstride+j-2]=new_data[i+rowstride+j+1]=new_data[i+rowstride+j+4]=old_data[x+y+1];
  new_data[i-rowstride+j-1]=new_data[i-rowstride+2]=new_data[i-rowstride+5]=new_data[i+j-1]=new_data[i+j+2]=new_data[i+j+5]=new_data[i+rowstride+j-1]=new_data[i+rowstride+j+2]=new_data[i+rowstride+j+5]=old_data[x+y+2];
}


///////////////////////////////////////////////////


// Rendered effects


gboolean  
do_effect(gchar *fx, gchar *params) {
  // returns FALSE if the user cancelled
  gint oundo_start=cfile->undo_start;
  gint oundo_end=cfile->undo_start;
  gchar *action;
  gchar effectstring[128];
  gdouble old_pb_fps=cfile->pb_fps;

  gchar *text=g_strdup_printf("%s frames %d to %d...",effectname,cfile->start,cfile->end);
  // create the default parameters which every function gets
  gchar *pdefault=g_strdup_printf ("%s %d %d",cfile->handle,cfile->start,cfile->end);
  // and append params
  gchar *fxcommand=g_strconcat ("smogrify ",fx," ",pdefault," ",params,NULL);
  
  g_free (pdefault);
  d_print(text);
  g_free(text);

  cfile->redoable=FALSE;
  cfile->undoable=FALSE;

  cfile->progress_start=cfile->undo_start=cfile->start;
  cfile->progress_end=cfile->undo_end=cfile->end;

  gtk_widget_set_sensitive (mainw->redo, FALSE);
  gtk_widget_set_sensitive (mainw->undo, FALSE);
  cfile->undo_action=UNDO_EFFECT;

  mainw->effects_paused=FALSE;
  unlink(cfile->info_file);
  cfile->pb_fps=1000000;

  if (mainw->internal_messaging==FALSE) {
    system(fxcommand);
  }
  else {
    if (mainw->rte&EFFECT_BLEND) {
      mainw->files[mainw->blend_file]->frameno=0;
    }
  }
  g_free (fxcommand);

  g_snprintf (effectstring,128,"%s frames %d to %d",effectname,cfile->undo_start,cfile->undo_end);

  if (do_progress_dialog(TRUE,TRUE,effectstring)==FALSE) {
    cfile->undo_start=oundo_start;
    cfile->undo_end=oundo_end;
    cfile->pb_fps=old_pb_fps;
    mainw->internal_messaging=FALSE;
    return FALSE;
  }

  cfile->changed=TRUE;
  gtk_widget_set_sensitive (mainw->undo, TRUE);
  cfile->undoable=TRUE;
  cfile->pb_fps=old_pb_fps;
  mainw->internal_messaging=FALSE;

  gtk_widget_set_sensitive (mainw->select_last, TRUE);

  action=g_strdup_printf("%s%s",g_ascii_strup(fx,1),fx+1);
  set_undoable (action,TRUE);
  g_free (action);
  switch_to_file (mainw->current_file,mainw->current_file);
  d_print("done\n");
  return TRUE;
}




void
on_monochrome_activate               (GtkMenuItem     *menuitem,
				      gpointer         user_data)
{
  gchar *fx=g_strdup ("monochrome");
  g_snprintf(effectname,256,"Applying monochrome to");
  do_effect(fx,NULL);
  g_free(fx);
}

void
on_blank_frames_activate               (GtkMenuItem     *menuitem,
				      gpointer         user_data)
{
  gchar *fx=g_strdup ("blank_frames");
  g_snprintf(effectname,256,"Blanking");
  do_effect(fx,NULL);
  g_free(fx);
}


void
on_charcoal_activate               (GtkMenuItem     *menuitem,
				    gpointer         user_data)
{
  gchar *params,*fx;
  g_snprintf(effectname,256,"Applying charcoal to");
  fx=g_strdup ("charcoal");
  params=g_strdup_printf("%d",2);
  do_effect(fx,params);
  g_free(params);
  g_free(fx);
}

void
on_edge_detect_activate               (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *fx=g_strdup ("edge");
  g_snprintf(effectname,256,"Edge detecting");
  do_effect(fx,NULL);
  g_free (fx);
}


void
on_colour_cycle_activate              (GtkMenuItem     *menuitem,
				       gpointer         user_data)
{
  gchar *fx=g_strdup ("cycle");
  g_snprintf(effectname,256,"Colour cycling");
  do_effect(fx,NULL);
  g_free (fx);
}

void
on_randomzoom_activate              (GtkMenuItem     *menuitem,
				     gpointer         user_data)
{
  gchar *params,*fx;
  g_snprintf(effectname,256,"Random zooming");
  fx=g_strdup ("randomzoom");
  params=g_strdup_printf("%d %d",cfile->vsize,cfile->hsize);
  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}


void
on_shrink_activate              (void)
{
  gchar *params,*fx;
  gint cx=cfile->hsize/2,cy=cfile->vsize/2;

  g_snprintf(effectname,256,"Shrinking/Expanding");
  fx=g_strdup ("shrink_expand");
  if (mainw->fx1_bool==TRUE) {
    // top
    cy=0;
  }
  else if (mainw->fx2_bool==TRUE) {
    // left
    cx=0;
  }
  else if (mainw->fx4_bool==TRUE) {
    // right
    cx=cfile->hsize;
  }
  else if (mainw->fx5_bool==TRUE) {
    // bottom
    cy=cfile->vsize;
  }
  else if (mainw->fx6_bool==TRUE) {
    // other (user defined)
    cx=(gint) mainw->fx3_val;
    cy=(gint) mainw->fx4_val;
  }

  params=g_strdup_printf("%d %d %.2f %.2f %d %d",cfile->vsize,cfile->hsize,mainw->fx1_val,mainw->fx2_val,cx,cy);
  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}


void
on_colorize_activate                  (gint type)
{
  gchar *params,*fx;
  gint cstart=(mainw->fx1_start*256+mainw->fx2_start)*256+mainw->fx3_start;
  gint cend=(mainw->fx1_end*256+mainw->fx2_end)*256+mainw->fx3_end;

  if (type==1) {
    g_snprintf(effectname,256,"Colorizing");
    fx=g_strdup ("colorize");
  }
  else {
    g_snprintf(effectname,256,"Colour Filter");
    fx=g_strdup ("colour_filter");
  }
  params=g_strdup_printf("%d %d %d %d %d",cstart,(int)mainw->fx1_val,cend,(int)mainw->fx2_val,mainw->fx1_bool);
  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}


void
on_negate_activate                    (void)
{
  gchar *params,*fx;
  g_snprintf(effectname,256,"Negating");
  fx=g_strdup ("negate");
  params=g_strdup_printf("%d %d",(int)mainw->fx1_val,(int)mainw->fx2_val);
  do_effect(fx,params);
  g_free (params);
  g_free (fx);
}

void
on_spin_activate                     (void)
{

  gchar *params,*fx;
  if (cfile->end==cfile->start) return;
  g_snprintf(effectname,256,"Spinning");
  fx=g_strdup ("spin");
  if (mainw->fx1_bool==FALSE) {
    params=g_strdup_printf("%d %d 0 %.4f",cfile->vsize,cfile->hsize,(mainw->fx2_val/(cfile->end-cfile->start)));
  } else {
    params=g_strdup_printf("%d %d 0 %.4f",cfile->vsize,cfile->hsize,mainw->fx1_val);
  }
  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}

void
on_shift_activate                     (gint type)
{

  gchar *params,*fx;
  g_snprintf(effectname,256,"Shifting");
  fx=g_strdup ("shift");
  if (type==1) {
    // shift x
    if (mainw->fx1_bool==FALSE) {
      //fixed
      params=g_strdup_printf("%d %d %d 0 0 0",cfile->vsize,cfile->hsize,(int)mainw->fx2_val);
    } else {
      //sliding
      params=g_strdup_printf("%d %d 0 0 %.2f 0 %d",cfile->vsize,cfile->hsize,mainw->fx1_val,(gint)mainw->fx2_bool);
    }
  }
  else if (type==2) {
    //shift y
    if (mainw->fx1_bool==FALSE) {
      //fixed
      params=g_strdup_printf("%d %d 0 %d 0 0 %d",cfile->vsize,cfile->hsize,(int)mainw->fx2_val,(gint)mainw->fx2_bool);
    } else {
      //sliding
      params=g_strdup_printf("%d %d 0 0 0 %.2f %d",cfile->vsize,cfile->hsize,mainw->fx1_val,(gint)mainw->fx2_bool);
    }
  }
  else {
    // combined - not implemented by GUI yet
    params=g_strdup_printf ("%d %d 0 0 1 1 %d",cfile->vsize,cfile->hsize,(gint)mainw->fx2_bool);
  }

  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}


void
on_swirl_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *params,*fx;
  g_snprintf(effectname,256,"Swirling");
  fx=g_strdup ("swirl");
  params=g_strdup_printf("%d",(int)mainw->fx1_val);
  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}

void
on_jumble_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *params,*fx;
  g_snprintf(effectname,256,"Jumbling");
  fx=g_strdup ("jumble");
  params=g_strdup_printf("%d",(int)mainw->fx1_val);
  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}

void
on_tunnel_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *params,*fx;
  g_snprintf(effectname,256,"Tunnelling");
  fx=g_strdup ("tunnel");
  params=g_strdup_printf("%d %d",cfile->vsize,cfile->hsize);
  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}

void
on_wave_activate                     (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *params,*fx;
  g_snprintf(effectname,256,"Adding waves to");
  fx=g_strdup ("wave");
  params=g_strdup_printf("%d",cfile->hsize);
  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}

void
on_flop_activate                     (GtkMenuItem     *menuitem,
				      gpointer         user_data)
{
  gchar *fx=g_strdup ("flop");
  g_snprintf(effectname,256,"Applying horizontal flip to");
  do_effect(fx,NULL);
  g_free (fx);
}

void
on_flip_activate                     (GtkMenuItem     *menuitem,
				      gpointer         user_data)
{
  gchar *fx=g_strdup ("flip");
  g_snprintf(effectname,256,"Applying vertical flip to");
  do_effect(fx,NULL);
  g_free (fx);
}


void
on_solarize_activate                     (void)
{
  gchar *params,*fx;
  g_snprintf(effectname,256,"Solarizing");
  fx=g_strdup ("solarize");
  params=g_strdup_printf("%d %d",(int)mainw->fx1_val,(int)mainw->fx2_val);
  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}


void
on_enhance_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *fx=g_strdup ("enhance");
  g_snprintf(effectname,256,"Enhancing");
  do_effect(fx,NULL);
  g_free (fx);
}

void
on_normalize_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *fx=g_strdup ("normalize");
  g_snprintf(effectname,256,"Normalizing");
  do_effect(fx,NULL);
  g_free (fx);
}

void
on_dream_activate                   (GtkMenuItem     *menuitem,
                                        gpointer         user_data)
{
  gchar *params,*fx;
  g_snprintf(effectname,256,"Dreamifying");
  fx=g_strdup ("dream");
  params=g_strdup_printf("%d %d",cfile->vsize,cfile->hsize);
  do_effect(fx,params);
  g_free(params);
  g_free (fx);
}




void
on_rotate_all_activate                     (void)
{
  // slightly more complicated because it is really a tool not an effect
  gchar *params=NULL,*fx;
  gint ostart=cfile->start;
  gint oend=cfile->end;
  int degrees;
  gchar *msg;
  gint current_file=mainw->current_file;

  cfile->ohsize=cfile->hsize;
  cfile->ovsize=cfile->vsize;

  // here we invalidate cfile->ohsize, cfile->ovsize
  if (cfile->undo_action==UNDO_TRIM_FRAMES||cfile->undo_action==UNDO_RESIZE||(cfile->undo_action==UNDO_ROTATE_ALL&&cfile->undo3_boolean==FALSE)) {
    set_undoable(NULL,FALSE);
  }

  if (mainw->fx1_bool||mainw->fx2_bool) {
    // +- 90, size will change...
    cfile->nokeep=TRUE;
    cfile->hsize+=cfile->vsize;
    cfile->vsize=cfile->hsize-cfile->vsize;
    cfile->hsize-=cfile->vsize;
  }
  
  g_snprintf(effectname,256,"Rotating");
  fx=g_strdup ("rotate");
  
  if (mainw->fx1_bool==TRUE) {
    degrees=90;
  }
  else if (mainw->fx2_bool==TRUE) {
    degrees=270;
  }
  else if (mainw->fx3_bool==TRUE) {
    degrees=180;
  }
  else {
    if ((degrees=mainw->fx1_val)==0.) return;
  }
  params=g_strdup_printf ("%d %d %d 0",cfile->vsize,cfile->hsize,degrees);

  msg=g_strdup_printf ("Rotating clip by %d degrees.\n",degrees);
  d_print (msg);
  g_free (msg);

  cfile->start=1;
  cfile->end=cfile->frames;

  if (do_effect(fx,params)==FALSE) {
    cfile->hsize=cfile->ohsize;
    cfile->vsize=cfile->ovsize;
    g_free(params);
    g_free (fx);
    
    cfile->start=ostart;
    cfile->end=oend;
    cfile->nokeep=FALSE;
    resize (1);
    return;
  }

  g_free(params);
  g_free (fx);

  cfile->start=ostart;
  cfile->end=oend;
  cfile->nokeep=FALSE;

  cfile->undo_action=UNDO_ROTATE_ALL;
  cfile->undo1_boolean=mainw->fx1_bool;
  cfile->undo2_boolean=mainw->fx2_bool;
  cfile->undo3_boolean=mainw->fx3_bool;

  // force a resize
  switch_to_file((mainw->current_file=0),current_file);
}


void
on_trim_frames_activate                     (void)
{
  gchar *params=NULL,*fx;
  gint ostart=cfile->start;
  gint oend=cfile->end;
  gchar *msg;
  gint current_file=mainw->current_file;

  cfile->ohsize=cfile->hsize;
  cfile->ovsize=cfile->vsize;

  // here we invalidate cfile->ohsize, cfile->ovsize
  if (cfile->undo_action==UNDO_TRIM_FRAMES||cfile->undo_action==UNDO_RESIZE||(cfile->undo_action==UNDO_ROTATE_ALL&&cfile->undo3_boolean==FALSE)) {
    set_undoable(NULL,FALSE);
  }

  g_snprintf(effectname,256,"Trimming frames");
  fx=g_strdup ("trim_frames");
  params=g_strdup_printf("%d %d %d %d %d %d",cfile->vsize,cfile->hsize,(gint)mainw->fx1_val,(gint)mainw->fx2_val,(gint)mainw->fx4_val,(gint)mainw->fx3_val);

  msg=g_strdup_printf ("Coordinates were: %d %d %d %d\n",(int)mainw->fx1_val,(int)mainw->fx2_val,(int)mainw->fx3_val,(int)mainw->fx4_val);
  d_print (msg);
  g_free (msg);

  cfile->start=1;
  cfile->end=cfile->frames;
  cfile->nokeep=TRUE;

  if (do_effect(fx,params)==FALSE) {
    g_free(params);
    g_free (fx);
    
    cfile->start=ostart;
    cfile->end=oend;
    cfile->nokeep=FALSE;
    return;
  }

  g_free(params);
  g_free (fx);

  cfile->nokeep=FALSE;
  cfile->start=ostart;
  cfile->end=oend;

  cfile->hsize=mainw->fx3_val;
  cfile->vsize=mainw->fx4_val;

  cfile->undo_action=UNDO_TRIM_FRAMES;
  cfile->undo1_int=mainw->fx1_val;
  cfile->undo2_int=mainw->fx2_val;
  cfile->undo3_int=mainw->fx3_val;
  cfile->undo4_int=mainw->fx4_val;

  // mark file size as unknown
  cfile->f_size=0;

  // force a resize
  switch_to_file ((mainw->current_file=0),current_file);
}











// realtime fx


  

gint 
realfx_progress (gboolean reset) {
  static int i;
  GError *error=NULL;
  gchar fname[256];
  gchar oname[256];
  GdkPixbuf *image;
  gchar *com;

  // this is called periodically from do_processing_dialog for internal effects

  if (reset==TRUE) {
    i=cfile->start;
    clear_mainw_msg();
    return 1;
  }

  if (mainw->effects_paused) return 1;

  // sig_progress...
  g_snprintf (mainw->msg,256,"%d",i);
  // load, effect, save frame

  g_snprintf(fname,256,"%s/%s/%08d.jpg",prefs->tmpdir,cfile->handle,i);
  g_snprintf(oname,256,"%s/%s/%08d.mgk",prefs->tmpdir,cfile->handle,i);

  image=gdk_pixbuf_new_from_file(fname,&error);

  if (image==NULL) {
    g_snprintf (mainw->msg,256,"error|missing image %d",i);
    return 1;
  }

  on_rte_apply (mainw->rte,image);
  gdk_pixbuf_save (image, oname, "jpeg", &error,"quality", "100", NULL);
  gdk_pixbuf_unref (image);
  
  if (++i>cfile->end) {
    com=g_strdup_printf ("smogrify mv_mgk %s %d %d",cfile->handle,cfile->start,cfile->end);
    system (com);
    g_free (com);
    mainw->internal_messaging=FALSE;
  }
  return 1;
}



void
on_realfx_activate                   (GtkMenuItem     *menuitem,
				      gpointer         user_data)
{
  g_snprintf(effectname,256,"Applying Realtime FX to");
  mainw->internal_messaging=TRUE;
  mainw->progress_fn=&realfx_progress;
  realfx_progress (TRUE);

  do_effect("Realtime FX",NULL);

  invalidate_pixel_buffers();

}

void 
on_rte_apply (gint eff_mask, GdkPixbuf *pixbuf) {
// realtime effects
// TODO - make into gmodules, allow assignment of realtime effects to ctrl-keys

  guchar *new_pixel_data;
  guchar byte;
  int i,j;

  int rowstride=gdk_pixbuf_get_rowstride(pixbuf);
  int height=gdk_pixbuf_get_height(pixbuf);
  gint size=height*rowstride;

  // assume RGB 24 bit
  if (gdk_pixbuf_get_bits_per_sample(pixbuf)!=8||gdk_pixbuf_get_n_channels(pixbuf)!=3) return;
  new_pixel_data=gdk_pixbuf_get_pixels(pixbuf);

  // real time effects
  if (mainw->rte&EFFECT_MIRRORX) {
    for (i=0;i<size;i+=rowstride) {
      for (j=0;j<rowstride/2;j++) {
	new_pixel_data[i+rowstride-1-j]=new_pixel_data[i+j];
      }
    }
  }
    
  if (mainw->rte&EFFECT_MIRRORY) {
    gdouble mratio=1.;
    if (mainw->rte_keys==EFFECT_MIRRORY) {
      mratio=256./(0xff^mainw->blend_factor);
    }
    for (i=0;i<size*mratio/(mratio+1);i+=rowstride) {
      for (j=0;j<rowstride;j++) {
	new_pixel_data[size-(int)(i/mratio/rowstride+1)*rowstride+j]=new_pixel_data[i+j];
      }
    }
  }

  if (mainw->rte&EFFECT_ALPHA) {
    if (mainw->old_pixel_data==NULL) {
      mainw->old_pixel_data=g_malloc0(size);
    }
    for (i=0;i<size;i+=rowstride) {
      for (j=0;j<rowstride;j++) {
	byte=new_pixel_data[i+j];
	new_pixel_data[i+j]=((char)byte+(char)mainw->old_pixel_data[i+j])/2;
	mainw->old_pixel_data[i+j]=byte;
      }
    }
  }
    
  if (mainw->rte&EFFECT_NOISE) {
    for (i=0;i<size;i+=rowstride) {
      for (j=0;j<rowstride;j++) {
	new_pixel_data[i+j]+=((fastrand()&0xF8000000)>>27)-16;
      }
    }
  }
    
  if (mainw->rte&EFFECT_NEGATE) {
    for (i=0;i<size;i+=rowstride) {
      for (j=0;j<rowstride;j++) {
	new_pixel_data[i+j]=new_pixel_data[i+j]^0xFF;
	;
      }
    }
  }
    
  if (mainw->rte&EFFECT_EDGE) {
    guint myluma;
    guint threshold=10000;
    gint nbr;
    int k,l;

    if (mainw->old_pixel_data==NULL) {
      mainw->old_pixel_data=g_malloc0(size);
    }
    memcpy(mainw->old_pixel_data,new_pixel_data,size);
    for (i=rowstride;i<size-rowstride;i+=rowstride) {
      for (j=3;j<rowstride-3;j+=3) {
	myluma=calc_luma (&mainw->old_pixel_data[i+j]);
	nbr=0;
	for (k=i-rowstride;k<=i+rowstride;k+=rowstride) {
	  for (l=j-3;l<=j+3;l+=3) {
	    if ((k!=i||j!=l)&&ABS (calc_luma (&mainw->old_pixel_data[k+l])-myluma)>threshold) nbr++;
	  }
	}
	if (nbr<2||nbr>5) {
	  nine_fill (new_pixel_data,i,j,mainw->old_pixel_data,i,j,rowstride);
	}
	else {
	  if (myluma<12500) {
	    make_black (&new_pixel_data[i+j]);
	  }
	  else{
	    if (myluma>20000) {
	      make_white (&new_pixel_data[i+j]);
	    }
	  }
	}
      }
    }
  }
  
  if (mainw->rte&EFFECT_POSTER) {
    for (i=0;i<size;i+=rowstride) {
      for (j=0;j<rowstride;j++) {
	new_pixel_data[i+j]=new_pixel_data[i+j]>128?255:0;
      }
    }
  }
    
  if (mainw->rte&EFFECT_REMOVE_BG) {
    // TODO - use HSI palette
    if (rowstride==cfile->hsize*3) {
      guint luma;
      guint av_luma;
      gint cpos;
      guint luma_threshold;
      luma_threshold=(mainw->rte_keys==EFFECT_REMOVE_BG?mainw->blend_factor*39:6000);

      if (mainw->av_luma_data==NULL) {
	mainw->av_luma_data=(guint *)g_malloc0(cfile->hsize*cfile->vsize*3*sizint);
	mainw->av_count=0;
      }
      
      mainw->av_count++;
      
      for (i=0;i<size;i+=rowstride) {
	cpos=i/rowstride*cfile->hsize*3;
	for (j=0;j<cfile->hsize*3;j+=3) {
	  luma=calc_luma (&new_pixel_data[i+j]);
	  if (mainw->av_count==1) {
	    mainw->av_luma_data[cpos+j/3]=luma;
	  }
	  av_luma=(guint)((gdouble)(mainw->av_luma_data[cpos+j/3]*mainw->av_count)+(gdouble)luma)/((gdouble)mainw->av_count+1.);
	  mainw->av_luma_data[cpos+j/3]=av_luma;
	  if (ABS (luma-av_luma)>(luma_threshold)) {
	    switch (mainw->rte_mode*(mainw->rte_keys==EFFECT_REMOVE_BG)) {
	    case 1:
	      // fire-ish effect
	      new_pixel_data[i+j]=((guchar)(((fastrand()>>24)+(new_pixel_data[i+j+1]=((guchar) (fastrand()>>25))))&0xff));     //R & G
	      new_pixel_data[i+j+2]=(guchar)0;                       //B
	      break;
	    //
	    case 2:
	      break;
	    case 0:
	      // make moving things black
	      make_black (&new_pixel_data[i+j]);
	    }
	  }
	  else {
	    // blue glow
	    if (mainw->rte_mode==2&&mainw->rte_keys==EFFECT_REMOVE_BG) {
	      new_pixel_data[i+j]=new_pixel_data[i+j+1]=((guchar)(fastrand()>>24));                                                 //R&G
	      new_pixel_data[i+j+2]=(guchar)255;  //B
	    }}}}}
  }

  if (mainw->rte&EFFECT_BLEND) {
    if (mainw->current_file!=mainw->blend_file) {
      GError *error=NULL;
      GdkPixbuf *blendr, *blendp;
      guchar *blend_data;
      gint threshold=mainw->blend_factor*100;

      if (++mainw->files[mainw->blend_file]->frameno>mainw->files[mainw->blend_file]->frames) mainw->files[mainw->blend_file]->frameno=1;
      blendp=gdk_pixbuf_new_from_file(g_strdup_printf ("%s/%s/%08d.jpg",prefs->tmpdir,mainw->files[mainw->blend_file]->handle,mainw->files[mainw->blend_file]->frameno),&error);
      if (error==NULL) {
	blend_data=gdk_pixbuf_get_pixels ((blendr=gdk_pixbuf_scale_simple(blendp,cfile->hsize,cfile->vsize,GDK_INTERP_NEAREST)));
	for (i=0;i<size;i+=rowstride) {
	  for (j=0;j<rowstride;j+=3) {
	    // avg
	    switch (mainw->rte_mode*(mainw->rte_keys==EFFECT_BLEND)) {
	    case 0:
	      // chroma blend
	      new_pixel_data[i+j]=(guchar)((new_pixel_data[i+j]*mainw->blend_factor+blend_data[i+j]*(0xFF^mainw->blend_factor))>>8);
	      new_pixel_data[i+j+1]=(guchar)((new_pixel_data[i+j+1]*mainw->blend_factor+blend_data[i+j+1]*(0xFF^mainw->blend_factor))>>8);
	      new_pixel_data[i+j+2]=(guchar)((new_pixel_data[i+j+2]*mainw->blend_factor+blend_data[i+j+2]*(0xFF^mainw->blend_factor))>>8);
	      break;
	    case 1:
	      // threshold luma blend (background)
	      if (calc_luma (&new_pixel_data[i+j])<(threshold)) {
		new_pixel_data[i+j]=blend_data[i+j];
		new_pixel_data[i+j+1]=blend_data[i+j+1];
		new_pixel_data[i+j+2]=blend_data[i+j+2];
	      }
	      break;	
	    case 2:
	      // threshold luma blend (foreground)
	      if (calc_luma (&blend_data[i+j])>(threshold)) {
		new_pixel_data[i+j]=blend_data[i+j];
		new_pixel_data[i+j+1]=blend_data[i+j+1];
		new_pixel_data[i+j+2]=blend_data[i+j+2];
	      }
	      break;
	    }
	  }
	}
	gdk_pixbuf_unref(blendp);
	gdk_pixbuf_unref(blendr);
      }
      else {
	g_error_free(error);
      }
    }
  }
}


void 
invalidate_pixel_buffers (void) {
  // invalidate all pixel buffers
  // this should call [stop()]/start() for all fx plugins

  // we call this when we start playing, when we switch to a different sized clip

  if (!(mainw->old_pixel_data==NULL)) {
    g_free(mainw->old_pixel_data);
    mainw->old_pixel_data=NULL;
  }

  if (!(mainw->av_luma_data==NULL)) {
    g_free(mainw->av_luma_data);
    mainw->av_luma_data=NULL;
  }

  if (!(mainw->uyvy_frame==NULL)) {
    g_free(mainw->uyvy_frame);
    mainw->uyvy_frame=NULL;
  }
  if (mainw->rte&EFFECT_BLEND&&(mainw->files[mainw->blend_file]==NULL||!mainw->files[mainw->blend_file]->frames||!mainw->files[mainw->blend_file]->is_loaded)) {
    mainw->blend_file=mainw->current_file;
  }
}



// this is the callback which happens when a rte is keyed
void
on_rte_activate            (GtkMenuItem     *menuitem,
			    gpointer         user_data)
{
  // TODO - call stop/start fn's of fx plugins
  guint new_rte=GPOINTER_TO_UINT(user_data);
  if (new_rte==EFFECT_NONE) {
    // switch up/down keys to default (fps change)
    mainw->rte=mainw->rte_keys=mainw->last_grabable_effect=EFFECT_NONE;
  }
  else {
    mainw->rte^=new_rte;
    if (mainw->rte&new_rte) {
      // switch is ON
      switch (new_rte) {
	// TODO - use can_grab capability to see whether to set last_grabable
      case EFFECT_REMOVE_BG:
	if (mainw->av_luma_data!=NULL) {
	  g_free (mainw->av_luma_data);
	  mainw->av_luma_data=NULL;
	}
	// allow switch of up/down/x keys
	mainw->rte_mode=0;
	mainw->rte_mode_max=2;
	mainw->last_grabable_effect=EFFECT_REMOVE_BG;
	break;
      case EFFECT_BLEND:
	mainw->blend_file=mainw->current_file;
	
	// allow switch of up/down/x keys
	mainw->rte_mode=0;
	mainw->rte_mode_max=2;
	mainw->last_grabable_effect=EFFECT_BLEND;
	break;
      case EFFECT_MIRRORY:
	// allow switch of up/down/x keys
	mainw->rte_mode=0;
	mainw->rte_mode_max=0;
	mainw->last_grabable_effect=EFFECT_MIRRORY;
	break;
      }
    }
  }
  if (cfile->play_paused) {
    load_frame_image (cfile->frameno,cfile->last_frameno);
  }
}



void
on_rte_grabkeys_activate            (GtkMenuItem     *menuitem,
				     gpointer         user_data)
{
  // assign the keys to the last key-grabable effect 
  mainw->rte_keys=mainw->last_grabable_effect;
  mainw->blend_factor=50;
}

void
on_rte_switchmode_activate            (GtkMenuItem     *menuitem,
				       gpointer         user_data)
{
  // TODO - call the next_mode() fn in the 
  // mainw->rte_keys fx plugin
  mainw->rte_mode=((mainw->rte_mode==mainw->rte_mode_max)?0:++mainw->rte_mode);
}
